"""The core module init file."""

from experiment_lab.core.base_config import BaseConfig
from experiment_lab.core.base_experiment import BaseExperiment
from experiment_lab.core.base_analysis import BaseAnalysis
from experiment_lab.core.runner import run_experiment



"""Helper functions to analyze the results of experiments."""

import abc
import json
from textwrap import indent
from typing import Any, Dict, Tuple
import pandas as pd
from torch import Value
import wandb
import tqdm
from hydra.core.hydra_config import HydraConfig
import logging
from pathlib import Path
import glob

from experiment_lab.core.base_config import BaseConfig


logger = logging.getLogger(__name__)

api = None


def get_api_instance() -> wandb.Api:
    """Gets an instance of the wandb api.

    Returns:
        wandb.Api: The wandb api instance.
    """
    global api
    if api is None:
        api = wandb.Api()
    return api


class BaseAnalysis(abc.ABC):
    """A generic base experiment analysis class."""

    def __init__(self, cfg: BaseConfig) -> None:
        """The constructor for the BaseAnalysis class.

        Args:
            cfg (BaseConfig): The config to analyze with.
        """
        self.cfg = cfg

    def load_data(self) -> Tuple[pd.DataFrame, Dict[str, Dict[str, Any]]]:
        """Function to load the run data of an experiment.

        Raises:
            ValueError: If wandb config is not specified, raise a ValueError.
            ValueError: If load_from_output_dir is set to true without any outputs to load, raise a ValueError.

        Returns:
            Tuple[pd.DataFrame, Dict[str, Dict[str, Any]]]: The dataframe containing all the run data and the dict containing all the configs.
        """
        if not self.cfg.analysis.load_from_output_dir:
            if self.cfg.wandb is None:
                raise ValueError("Not wandb config specified")
            api = get_api_instance()
            wandb_runs = api.runs(
                path=self.cfg.wandb["project"],
                filters={
                    "$and": [
                        {"state": "finished"},
                        {"config.experiment_name": self.cfg.experiment_name},
                    ]
                    + self.cfg.analysis.filters,
                },
                include_sweeps=False,
            )

            full_df = []

            configs = {}

            for wandb_run in tqdm.tqdm(wandb_runs):
                if self.cfg.analysis.all_keys_per_step:
                    df = pd.DataFrame(
                        wandb_run.scan_history(keys=self.cfg.analysis.wandb_keys),
                    )
                else:
                    df = pd.DataFrame()
                    for key in self.cfg.analysis.wandb_keys:
                        temp_df = pd.DataFrame(
                            wandb_run.scan_history(keys=[key]),
                        )
                        df = pd.merge(
                            df, temp_df, left_index=True, right_index=True, how="outer"
                        )
                df["experiment_id"] = wandb_run.config["experiment_id"]
                df["run_id"] = wandb_run.id
                if wandb_run.config["experiment_id"] not in configs:
                    configs[wandb_run.config["experiment_id"]] = wandb_run.config
                if self.cfg.analysis.index is not None:
                    df.set_index(
                        ["experiment_id", "run_id", self.cfg.analysis.index],
                        inplace=True,
                    )
                else:
                    df.set_index(["experiment_id", "run_id"], append=True, inplace=True)
                    df = df.reorder_levels([1, 2, 0])

                full_df.append(df)

            full_df = pd.concat(full_df)
            full_df.sort_index(inplace=True)
        else:
            if self.cfg.analysis.output_dir_to_load_from is None:
                root_output_dir = Path(self.output_directory).parent.parent.absolute()
                glob_results = sorted(glob.glob(f"{root_output_dir}/*/*/run_data.pkl"))
                if len(glob_results) == 0:
                    raise ValueError(
                        "Cannot load results from output directory since there are no other analysis runs."
                    )
                output_dir_to_load_from = Path(glob_results[-1]).parent
            else:
                output_dir_to_load_from = self.cfg.analysis.output_dir_to_load_from
            full_df = pd.read_pickle(f"{output_dir_to_load_from}/run_data.pkl")
            with open(f"{output_dir_to_load_from}/configs.json", "r") as f:
                configs = json.load(f)

        full_df.to_pickle(f"{self.output_directory}/run_data.pkl")
        with open(f"{self.output_directory}/configs.json", "w") as f:
            json.dump(configs, f, indent=4)

        return full_df, configs

    def _analyze_wrapper(self) -> Any:
        """A wrapper for the analyze function.

        Returns:
            Any: The results of the analysis.
        """
        self.output_directory = HydraConfig.get().runtime.output_dir
        df, configs = self.load_data()
        results = self.analyze(df=df, configs=configs)
        logger.info(results)
        return results

    @abc.abstractmethod
    def analyze(self, df: pd.DataFrame, configs: Dict[str, Dict[str, Any]]) -> Any:
        """Runs the analysis for the experiment.

        Args:
            df (pd.DataFrame): The results dataframe from wandb.
            configs (Dict[str, Dict[str, Any]]): The configs for each experiment.

        Returns:
            Any: The results of the analysis.
        """
        pass



"""Main entrypoint to run generic rl experiments."""

import minigrid

from experiment_lab.core.runner import run_experiment, root_config_folder
from experiment_lab.experiments.rl.experiment import RLExperiment
from experiment_lab.experiments.rl.config import RLConfig, register_configs
from experiment_lab.experiments.rl.analysis import RLAnalysis

from functools import partial


run_rl_experiment = partial(
    run_experiment,
    experiment_cls=RLExperiment,
    config_cls=RLConfig,
    analysis_cls=RLAnalysis,
    register_configs=register_configs,
    config_path=f"{root_config_folder}/rl",
)


if __name__ == "__main__":
    run_rl_experiment()



"""Init file for the networks module."""

from experiment_lab.common.networks.network import (
    create_network,
    create_aggregated_network,
    create_multi_network,
)
from experiment_lab.common.networks.mlp import create_mlp_network



"""Init file for the rl experiments module."""

from experiment_lab.experiments.rl.experiment import RLExperiment
from experiment_lab.experiments.rl.config import RLConfig, register_configs
from experiment_lab.experiments.rl.main import run_rl_experiment



"""Main entrypoint to run generic monte carlo experiments."""

from experiment_lab.core.runner import run_experiment, root_config_folder
from experiment_lab.experiments.monte_carlo.experiment import MCExperiment
from experiment_lab.experiments.monte_carlo.config import MCConfig, register_configs

from functools import partial


run_mc_experiment = partial(
    run_experiment,
    experiment_cls=MCExperiment,
    config_cls=MCConfig,
    register_configs=register_configs,
    config_path=f"{root_config_folder}/mc",
)


if __name__ == "__main__":
    run_mc_experiment()



"""List resolvers for easily generating lists in config files"""

from typing import Any, Callable, Sequence
import numpy as np


def wrap_f(f: Callable):
    """Wraps function and converts output to list.

    Args:
        f (Callable): The function to wrap.
    """

    def g(*args, **kwargs) -> Sequence[Any]:
        return list(f(*args, **kwargs))

    return g


def dup(n: int, x: Any) -> Sequence[Any]:
    """Duplicates x into a list of length n.

    Args:
        n (int): The length of the resulting list.
        x (Any): The value to duplicate.

    Returns:
        Sequence[Any]: The list containing x, n times.
    """
    return [x for _ in range(n)]


def wrap_around(n: int, lst: Sequence[Any]) -> Sequence[Any]:
    """Generates a list of length n by iterating through lst and wrapping around if the values run out.

    Args:
        n (int): The length of the resulting list.
        lst (Sequence[Any]): The list to iterate through.

    Returns:
        Sequence[Any]: The list wrapping around lst.
    """
    return [lst[i % len(lst)] for i in range(n)]


def toggle(n: int, fst: Any = False, snd: Any = True) -> Sequence[Any]:
    """Toggles between two values until a list of length n is generated.

    Args:
        n (int): The length of the resulting list.
        fst (Any, optional): The first (and third, fifth, ...) value to use. Defaults to False.
        snd (Any, optional): The second (and fourth, sixth, ...) value to use. Defaults to True.

    Returns:
        Sequence[Any]: The resulting list toggling between the two values.
    """
    return wrap_around(n=n, lst=[fst, snd])


range = wrap_f(range)
arange = wrap_f(np.arange)
linspace = wrap_f(np.linspace)



"""The rl experiment python file."""

import os
from typing import Any, List, Type, Dict
import gymnasium as gym
import hydra
import torch
from wandb.integration.sb3 import WandbCallback

from stable_baselines3.common.callbacks import CallbackList
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.policies import BasePolicy
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.vec_env import VecVideoRecorder


from experiment_lab.core.base_experiment import BaseExperiment
from experiment_lab.experiments.rl.config import RLConfig
from experiment_lab.experiments.rl.environment import GeneralVecEnv
from experiment_lab.common.utils import default


class RLExperiment(BaseExperiment):
    """The rl experiment class."""

    def __init__(self, cfg: RLConfig) -> None:
        """Constructor for rl experiment.

        Args:
            cfg (RLConfig): The config to run the experiment with.
        """
        self.cfg = cfg
        self.initialize_experiment()

    def initialize_experiment(self) -> None:
        """Initializes the rl experiment"""
        super().initialize_experiment()

        self.wrapper_cls_lst: List[Type[gym.Wrapper]] = (
            []
            if self.cfg.wrapper_cls_lst is None
            else [
                hydra.utils.get_class(wrapper_cls)
                for wrapper_cls in self.cfg.wrapper_cls_lst
            ]
        )
        self.policy_cls: Type[BasePolicy] = hydra.utils.get_class(self.cfg.policy_cls)
        self.callback_cls_lst: List[Type[BaseCallback]] = (
            []
            if self.cfg.callback_cls_lst is None
            else [
                hydra.utils.get_class(callback_cls)
                for callback_cls in self.cfg.callback_cls_lst
            ]
        )

        self.wrapper_kwargs_lst: List[Dict[str, Any]] = default(
            self.cfg.wrapper_kwargs_lst, []
        )
        for _ in range(len(self.wrapper_kwargs_lst), len(self.wrapper_cls_lst)):
            self.wrapper_kwargs_lst.append({})

        self.callback_kwargs_lst: List[Dict[str, Any]] = default(
            self.cfg.callback_kwargs_lst, []
        )
        for _ in range(len(self.callback_kwargs_lst), len(self.callback_cls_lst)):
            self.callback_kwargs_lst.append({})

        self.model_kwargs: Dict[str, Any] = default(self.cfg.model_kwargs, {})

        self.transfer_steps: List[int] = default(self.cfg.transfer_steps, [])

        self.device: torch.device = torch.device(self.cfg.device)

        self.additional_wandb_init_kwargs["sync_tensorboard"] = True
        self.additional_wandb_init_kwargs["save_code"] = True
        if self.cfg.record_policy_videos:
            self.additional_wandb_init_kwargs["monitor_gym"] = True

    def single_run(
        self, run_id: str, run_output_path: str, seed: int | None = None
    ) -> Any:
        """A single run of the rl experiment.

        Args:
            run_id (str): The run id.
            run_output_path (str): The run output path.
            seed (int | None, optional): The initial seed.. Defaults to None.

        Returns:
            Any: Any resulting metrics.
        """

        # Setup the device to train on
        if self.cfg.gpu_idx is not None:
            os.environ["CUDA_VISIBLE_DEVICES"] = str(self.cfg.gpu_idx)

        # Create the environment
        env = GeneralVecEnv(
            env_config=self.cfg.env_config,
            transfer_steps=self.cfg.transfer_steps,
            wrappers=self.wrapper_cls_lst,
            wrapper_kwargs_lst=self.cfg.wrapper_kwargs_lst,
            n_envs=self.cfg.n_envs,
            seed=seed,
            monitor_dir=self.cfg.monitor_dir,
            monitor_kwargs=self.cfg.monitor_kwargs,
            start_method=self.cfg.start_method,
            render_mode=(
                "rgb_array" if self.cfg.record_policy_videos else self.cfg.render_mode
            ),
        )
        if self.cfg.record_policy_videos:
            env = VecVideoRecorder(
                env,
                f"./wandb/videos/{run_id}",
                record_video_trigger=lambda x: x
                % (self.cfg.video_freq // self.cfg.n_envs)
                == 0,
                video_length=self.cfg.video_length,
                name_prefix=self.cfg.video_name_prefix,
            )

        env.seed(seed)
        env.reset()

        # Setup the model
        model = hydra.utils.instantiate(
            {
                "_target_": self.cfg.model_cls,
            },
            env=env,
            policy=self.policy_cls,
            policy_kwargs=self.cfg.policy_kwargs,
            device=self.device,
            tensorboard_log=f"./logs/" if self.cfg.log else None,
            verbose=self.cfg.verbose,
            **self.model_kwargs,
        )
        model.set_random_seed(seed)
        model_save_path = f"{run_output_path}/models/"

        # Instantiate the callbacks
        callback_instances = []
        if (
            self.cfg.log
            and self.cfg.wandb
            and self.cfg.wandb_callback_kwargs is not None
        ):
            callback_instances.append(
                WandbCallback(
                    gradient_save_freq=self.cfg.wandb_callback_kwargs.gradient_save_freq,
                    model_save_freq=self.cfg.wandb_callback_kwargs.model_save_freq,
                    model_save_path=model_save_path,
                    verbose=self.cfg.verbose,
                )
            )
        for callback_cls, callback_kwargs in zip(
            self.callback_cls_lst, self.callback_kwargs_lst
        ):
            callback_instances.append(
                callback_cls(**hydra.utils.instantiate(callback_kwargs))
            )

        # Run the algorithm
        model.learn(
            total_timesteps=self.cfg.total_time_steps,
            log_interval=self.cfg.log_interval,
            tb_log_name=run_id,
            callback=(
                CallbackList(callback_instances)
                if len(callback_instances) > 0
                else None
            ),
        )

        env.close()

        if self.cfg.save_model:
            os.makedirs(model_save_path, exist_ok=True)
            model.save(f"{model_save_path}/final.zip")



"""Init file for the components module"""

from experiment_lab.experiments.monte_carlo.components.aggregators import (
    BaseAggregator,
    NpAggregator,
    MultipleAggregators,
)
from experiment_lab.experiments.monte_carlo.components.samplers import (
    BaseSampler,
    UniformSampler,
)
from experiment_lab.experiments.monte_carlo.components.sample_filters import (
    BaseSampleFilter,
    PassThroughSampleFilter,
)
from experiment_lab.experiments.monte_carlo.components.post_processors import (
    BasePostProcessor,
    PassThroughPostProcessor,
)



"""The supervised learning experiment file."""

from json import load
import logging
from typing import Any, Dict
import os
import hydra
import torch
from torch import nn, optim
import torch.utils.data
import wandb

from experiment_lab.common.utils import default
from experiment_lab.core.base_experiment import BaseExperiment
from experiment_lab.experiments.supervised.config import IntervalType, SupervisedConfig


logger = logging.getLogger(__name__)


class SupervisedExperiment(BaseExperiment):
    """The supervised learning experiment class."""

    def __init__(self, cfg: SupervisedConfig) -> None:
        """The constructor for the supervised learning experiment."""
        self.cfg = cfg
        self.initialize_experiment()

    def initialize_experiment(self) -> None:
        """Initializes the supervised experiment config."""
        super().initialize_experiment()

        self.dataset: torch.utils.data.Dataset = hydra.utils.instantiate(
            self.cfg.dataset
        )
        self.device: torch.device = torch.device(self.cfg.device)
        self.data_loader_kwargs: Dict[str, Any] = default(
            self.cfg.data_loader_kwargs, {}
        )

    def apply_model_on_epoch(
        self,
        model: nn.Module,
        loader: torch.utils.data.DataLoader,
        optimizer: optim.Optimizer,
        scheduler: optim.lr_scheduler.LRScheduler | None,
        loss: nn.Module,
        train_with: bool = False,
        epoch: int = 0,
    ):
        model.train()
        losses = []
        num_batches = len(loader)
        for batch_idx, (input_data, target_data) in enumerate(loader):
            batch_idx += 1
            input_data, target_data = (
                input_data.to(self.device),
                target_data.to(self.device),
            )
            if train_with:
                optimizer.zero_grad()
                output_data = model(input_data)
                loss_val = loss(output_data, target_data)
                loss_val.backward()
                optimizer.step()
                if scheduler is not None:
                    scheduler.step()
                if (
                    (
                        self.cfg.logging_schedule.log_type == IntervalType.batches
                        or self.cfg.logging_schedule.log_type
                        == IntervalType.batches_and_epochs
                    )
                    and self.cfg.logging_schedule.log_freq > 0
                    and batch_idx % self.cfg.logging_schedule.log_freq == 0
                ):
                    logging.info(
                        f"epoch {epoch} / {self.cfg.num_epochs}; batch {batch_idx} / {num_batches} - train loss: {loss_val.item()};"
                    )
                if (
                    (
                        self.cfg.logging_schedule.wandb_log_type == IntervalType.batches
                        or self.cfg.logging_schedule.wandb_log_type
                        == IntervalType.batches_and_epochs
                    )
                    and self.cfg.logging_schedule.wandb_log_freq > 0
                    and batch_idx % self.cfg.logging_schedule.wandb_log_freq == 0
                ):
                    wandb.log(
                        {
                            "epoch": epoch,
                            "batch_idx": batch_idx,
                            "train_loss": loss_val.item(),
                        }
                    )
                losses.append(loss_val.item())
            else:
                with torch.no_grad():
                    output_data = model(input_data)
                    loss_val = loss(output_data, target_data)
                    losses.append(loss_val.item())
        return torch.Tensor(losses).to(self.device).mean()

    def single_run(
        self, run_id: str, run_output_path: str, seed: int | None = None
    ) -> Any:
        """A single run of the supervised learning experiment.

        Args:
            run_id (str): The run id.
            run_output_path (str): The run output path.
            seed (int | None, optional): The initial random seed. Defaults to None.

        Returns:
            Any: The post processed results.
        """
        # Setup the device to train on
        if self.cfg.gpu_idx is not None:
            os.environ["CUDA_VISIBLE_DEVICES"] = str(self.cfg.gpu_idx)

        if seed is not None:
            torch.manual_seed(seed=seed)

        # Setting up data
        train_dataset, validation_dataset, test_dataset = torch.utils.data.random_split(
            dataset=self.dataset,
            lengths=[self.cfg.train_ratio, self.cfg.val_ratio, self.cfg.test_ratio],
        )

        train_dataloader, validation_dataloader, test_dataloader = (
            torch.utils.data.DataLoader(
                dataset=train_dataset, **self.data_loader_kwargs
            ),
            torch.utils.data.DataLoader(
                dataset=validation_dataset, **self.data_loader_kwargs
            ),
            torch.utils.data.DataLoader(
                dataset=test_dataset, **self.data_loader_kwargs
            ),
        )

        # Setting up model
        model: nn.Module = hydra.utils.instantiate(self.cfg.model).to(self.device)
        if self.cfg.load_model_path is not None:
            model.load_state_dict(torch.load(self.cfg.load_model_path))

        # Setting up optimizer
        optimizer: optim.Optimizer = hydra.utils.instantiate(
            self.cfg.optimizer, model.parameters()
        )
        scheduler: optim.lr_scheduler.LRScheduler | None = (
            None
            if self.cfg.scheduler is None
            else hydra.utils.instantiate(self.cfg.scheduler, optimizer)
        )

        # Setting up loss
        loss = hydra.utils.instantiate(self.cfg.loss)

        os.makedirs(f"{run_output_path}/models", exist_ok=True)

        for epoch in range(0, self.cfg.num_epochs + 1):
            train_loss = self.apply_model_on_epoch(
                model=model,
                loader=train_dataloader,
                optimizer=optimizer,
                scheduler=scheduler,
                loss=loss,
                train_with=epoch > 0,
                epoch=epoch,
            )
            validation_loss = self.apply_model_on_epoch(
                model=model,
                loader=validation_dataloader,
                optimizer=optimizer,
                scheduler=scheduler,
                loss=loss,
                train_with=False,
                epoch=epoch,
            )
            test_loss = self.apply_model_on_epoch(
                model=model,
                loader=test_dataloader,
                optimizer=optimizer,
                scheduler=scheduler,
                loss=loss,
                train_with=False,
                epoch=epoch,
            )

            if self.cfg.save_model_freq > 0 and epoch % self.cfg.save_model_freq == 0:
                torch.save(
                    model.state_dict(),
                    f"{run_output_path}/models/model_epoch_{epoch}.pl",
                )

            if (
                (
                    self.cfg.logging_schedule.log_type == IntervalType.epochs
                    or self.cfg.logging_schedule.log_type
                    == IntervalType.batches_and_epochs
                )
                and self.cfg.logging_schedule.log_freq > 0
                and epoch % self.cfg.logging_schedule.log_freq == 0
            ):
                logging.info(
                    f"epoch {epoch} / {self.cfg.num_epochs}; - train loss: {train_loss.item()}; val_loss: {validation_loss.item()}; test_loss: {test_loss.item()}"
                )
            if (
                (
                    self.cfg.logging_schedule.wandb_log_type == IntervalType.epochs
                    or self.cfg.logging_schedule.wandb_log_type
                    == IntervalType.batches_and_epochs
                )
                and self.cfg.logging_schedule.wandb_log_freq > 0
                and epoch % self.cfg.logging_schedule.wandb_log_freq == 0
            ):
                wandb.log(
                    {
                        "epoch": epoch,
                        "train_loss": train_loss.item(),
                        "val_loss": validation_loss.item(),
                        "test_loss": test_loss.item(),
                    }
                )

        torch.save(
            model.state_dict(),
            f"{run_output_path}/models/final_model_epoch_{self.cfg.num_epochs}.pl",
        )



"""A python module containing functions to create generic networks."""

import collections.abc
from typing import Any, Callable, Dict, Sequence, Type, TypeVar
import hydra
import torch
from torch import nn

from experiment_lab.common.utils import default


def create_network(
    layer_cls: str | Type[nn.Module] | Sequence[Type[nn.Module] | str | None] | None,
    n_layers: int,
    layer_kwargs: Dict[str, Any] | Sequence[Dict[str, Any] | None] | None = None,
    constant_layer_kwargs: Dict[str, Any] | None = None,
    layer_activations: nn.Module | Sequence[nn.Module | None] | None = None,
    final_activation: nn.Module | None = None,
    dropout_p: Sequence[float | None] | float | None = None,
) -> nn.Module:
    """A generic create sequential network function

    Args:
        layer_cls (Type[nn.Module] | Sequence[Type[nn.Module]  |  None] | None): The layer class type(s) to use for each layer in the network.
        n_layers (int): The number of layers.
        layer_kwargs (Dict[str, Any] | Sequence[Dict[str, Any]  |  None] | None, optional): The kwargs to pass to the each layer. Defaults to None.
        constant_layer_kwargs (Dict[str, Any] | None, optional): The kwargs to pass to all the layers. Defaults to None.
        layer_activations (nn.Module | Sequence[nn.Module  |  None] | None, optional): The activation function to use after each layer. Defaults to None.
        final_activation (nn.Module | None, optional): The activation function to use at the end. Defaults to None.
        dropout_p (Sequence[float | None] | float | None, optional): The probability of dropout for each node. Defaults to None.

    Returns:
        nn.Module: The full network torch module.
    """

    network = nn.Sequential()

    constant_layer_kwargs_defaulted: Dict[str, Any] = default(constant_layer_kwargs, {})

    T = TypeVar("T")

    def get_from_lst(
        lst: T | Sequence[T | None] | None, i: int, default_value: T | None = None
    ) -> T | None:
        val: T | None = None
        if lst is None:
            val = default_value
        elif isinstance(lst, collections.abc.Sequence) and type(lst) != str:
            if i < len(lst):
                val = lst[i]
            else:
                val = lst[-1]
        else:
            val = lst  # type: ignore
        if val is None:
            return default_value
        else:
            return val

    for i in range(n_layers):
        layer_cls_i = get_from_lst(lst=layer_cls, i=i)
        if isinstance(layer_cls_i, str):
            layer_cls_i = hydra.utils.get_class(layer_cls_i)
        layer_kwargs_i = get_from_lst(lst=layer_kwargs, i=i, default_value={})
        dropout_p_i = get_from_lst(lst=dropout_p, i=i)
        activation = (
            final_activation
            if i == n_layers - 1
            else get_from_lst(lst=layer_activations, i=i, default_value=None)
        )
        if layer_cls_i is not None and layer_kwargs_i is not None:
            network.append(
                layer_cls_i(**layer_kwargs_i, **constant_layer_kwargs_defaulted)
            )
        if activation is not None:
            network.append(activation)
        if i < n_layers - 1 and dropout_p_i is not None and dropout_p_i > 0.0:
            network.append(nn.Dropout(p=dropout_p_i))

    return network


class AggregatedNetwork(nn.Module):
    """A class for aggregated multi input networks."""

    def __init__(
        self,
        module_lst: Sequence[nn.Module],
        aggregator: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        output_module: nn.Module,
    ) -> None:
        """The constructor for the multi input network.

        Args:
            module_lst (Sequence[nn.Module]): The list of modules to apply before aggregation.
            aggregator (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): The aggregator to use on the tensors.
            output_module (nn.Module): The module to apply after aggregation.
        """
        super().__init__()
        self.module_lst = module_lst
        self.aggregator = aggregator
        self.output_module = output_module

    def forward(self, xs: Sequence[torch.Tensor]) -> torch.Tensor:
        """The forward function of the aggregated network.

        Args:
            xs (Sequence[torch.Tensor]): The inputs to the network.

        Returns:
            torch.Tensor: The output of the network.
        """
        zs: Sequence[torch.Tensor] = [m(x) for x, m in zip(xs, self.module_lst)]
        agg_z = zs[0]
        for z in zs[1:]:
            agg_z = self.aggregator(agg_z, z)
        return self.output_module(agg_z)


class MultiNetwork(nn.Module):
    """A class for a multi output network."""

    def __init__(
        self,
        input_module: nn.Module,
        module_lst: Sequence[nn.Module],
    ) -> None:
        """Constructor for multi output network.

        Args:
            input_module (nn.Module): The module to apply on the input.
            module_lst (Sequence[nn.Module]): The modules to apply after input_module has been applied.
        """
        super().__init__()
        self.input_module = input_module
        self.module_lst = module_lst

    def forward(self, x: torch.Tensor) -> Sequence[torch.Tensor]:
        """The forward function of the multi output network

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            Sequence[torch.Tensor]: The outputs.
        """
        z = self.input_module(x)
        return [m(z) for m in self.module_lst]


def create_aggregated_network(
    module_lst: Sequence[nn.Module],
    aggregator: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    output_module: nn.Module,
) -> AggregatedNetwork:
    """Creates an instance of the complex multi input network.

        Args:
            module_lst (Sequence[nn.Module]): The list of modules to apply before aggregation.
            aggregator (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): The aggregator to use on the tensors.
            output_module (nn.Module): The module to apply after aggregation.

    Returns:
        AggregatedNetwork: The complex multi input network.
    """
    return AggregatedNetwork(module_lst, aggregator, output_module)


def create_multi_network(
    input_module: nn.Module,
    module_lst: Sequence[nn.Module],
) -> MultiNetwork:
    """Creates an instance of the multi output network.

    Args:
        input_module (nn.Module): The input module to apply.
        module_lst (Sequence[nn.Module]): The sequence of modules to apply after the input_module.

    Returns:
        MultiNetwork: The complex multi output network.
    """
    return MultiNetwork(input_module=input_module, module_lst=module_lst)



"""A python module containing a basic MLP network function."""

from typing import Any, Dict, Sequence, Type
from torch import nn
from experiment_lab.common.networks.network import create_network


def create_mlp_network(
    layer_sizes: Sequence[int],
    linear_layer_cls: Type[nn.Module] = nn.Linear,
    constant_layer_kwargs: Dict[str, Any] | None = None,
    layer_activations: nn.Module | Sequence[nn.Module | None] | None = None,
    final_activation: nn.Module | None = None,
    dropout_p: Sequence[float | None] | float | None = None,
) -> nn.Module:
    """A create multi layer percetron function.

    Args:
        layer_sizes (Sequence[int]): The layer sizes to use in the network.
        linear_layer_cls (Type[nn.Module], optional): The linear layer class to use. Defaults to nn.Linear.
        layer_activations (nn.Module | Sequence[nn.Module  |  None] | None, optional): The layer activation funcitons.. Defaults to None.
        final_activation (nn.Module | None, optional): The final activation function before the output. Defaults to None.
        dropout_p (Sequence[float  |  None] | float | None, optional): The probability of dropout per layer.. Defaults to None.

    Returns:
        nn.Module: The mlp network.
    """
    assert (
        len(layer_sizes) >= 2
    ), "Layer sizes must be at least 2 to include at least one layer in the network."
    return create_network(
        layer_cls=linear_layer_cls,
        n_layers=len(layer_sizes) - 1,
        layer_kwargs=[
            dict(in_features=layer_sizes[i], out_features=layer_sizes[i + 1])
            for i in range(len(layer_sizes) - 1)
        ],
        constant_layer_kwargs=constant_layer_kwargs,
        layer_activations=layer_activations,
        final_activation=final_activation,
        dropout_p=dropout_p,
    )



"""Python file containing the configs for rl experiments."""

from dataclasses import dataclass, field
from typing import Any, Dict, List
from hydra.core.config_store import ConfigStore

from experiment_lab.core.base_config import BaseConfig


@dataclass
class WandbCallbackConfig:
    """The wandb callback config dataclass."""

    model_save_freq: int = 1_000
    gradient_save_freq: int = 0


@dataclass
class RLConfig(BaseConfig):
    """The RL Config dataclass."""

    env_config: Dict[str, Any] = field(
        default_factory=lambda: {"env_id": "CartPole-v1"}
    )
    transfer_steps: List[int] | None = None

    total_time_steps: int = 10_000
    n_envs: int = 1

    wrapper_cls_lst: List[str] | None = None
    wrapper_kwargs_lst: List[Dict[str, Any]] | None = None

    model_cls: str = "stable_baselines3.PPO"
    model_kwargs: Dict[str, Any] | None = None

    policy_cls: str = "stable_baselines3.ppo.MlpPolicy"
    policy_kwargs: Dict[str, Any] | None = None

    callback_cls_lst: List[str] | None = None
    callback_kwargs_lst: List[Dict[str, Any]] | None = None
    wandb_callback_kwargs: WandbCallbackConfig | None = field(
        default_factory=WandbCallbackConfig
    )

    verbose: int = 1

    monitor_dir: str | None = None
    monitor_kwargs: Dict[str, Any] | None = None

    record_policy_videos: bool = False
    video_length: int = 200
    video_freq: int = 5000
    video_name_prefix: str = "agent"

    start_method: str | None = None
    render_mode: str | None = None

    log: bool = True
    log_interval: int = 1
    save_model: bool = True

    device: str = "mps"
    gpu_idx: int | None = None


def register_configs():
    """Registers the rl config."""
    cs = ConfigStore.instance()
    cs.store(name="wandb_callback_defaults", node=WandbCallbackConfig)
    cs.store(name="rl_config", node=RLConfig)



"""The setup code for the library."""

import pathlib

from setuptools import setup


CWD = pathlib.Path(__file__).absolute().parent


def get_version() -> str:
    """Pulls the version from the __init__ file in the library.

    Raises:
        RuntimeError: Raises when no __version__ was found.

    Returns:
        str: The current version of the library.
    """
    path = CWD / "experiment_lab" / "__init__.py"
    content = path.read_text()

    for line in content.splitlines():
        if line.startswith("__version__"):
            return line.strip().split()[-1].strip().strip('"')
    raise RuntimeError("bad version data in __init__.py")


def get_long_description() -> str:
    """Reads the long description from the README.md file.

    Returns:
        str: The long description.
    """

    with open("README.md") as fh:
        long_description = "".join(fh.readlines())
    return long_description


setup(
    name="experiment_lab",
    version=get_version(),
    long_description=get_long_description(),
)



"""File containing the main entrypoint function."""

import os

from typing import Any, Callable, Type, Sequence
import hydra
from omegaconf import DictConfig, OmegaConf

import experiment_lab
from experiment_lab.common.resolvers import register_resolvers
from experiment_lab.core.base_analysis import BaseAnalysis
from experiment_lab.core.base_config import BaseConfig, register_configs
from experiment_lab.core.base_experiment import BaseExperiment

root_config_folder = f"{os.path.dirname(experiment_lab.__file__)}/configs"


def run_experiment(
    experiment_cls: Type[BaseExperiment],
    config_cls: Type[BaseConfig] = BaseConfig,
    analysis_cls: Type[BaseAnalysis] | None = None,
    register_configs: Callable[[], None] = register_configs,
    register_resolvers: Callable[[], None] = register_resolvers,
    config_path: str = root_config_folder,
    config_name: str = "config",
) -> Any:
    """The main entrypoint to collect all the hydra config and run the experiment.

    Args:
        experiment_cls (Type[BaseExperiment]): The experiment class.
        config_cls (Type[BaseConfig], optional): The config class. Defaults to BaseConfig.
        register_configs (Callable[[], None], optional): The function to register any configs. Defaults to register_configs.
        register_resolvers (Callable[[], None], optional): The function to register any resolvers. Defaults to register_resolvers.
        config_path (str, optional): The config path. Defaults to "./configs".
        config_name (str, optional): The default config name. Defaults to "config".

    Returns:
        Sequence[Any]: The list of results from the runs.
    """
    register_resolvers()
    register_configs()

    config_path = os.path.join(os.getcwd(), config_path)

    @hydra.main(config_path=config_path, config_name=config_name, version_base=None)
    def main(dict_cfg: DictConfig) -> Any:
        OmegaConf.resolve(dict_cfg)
        cfg: config_cls = OmegaConf.to_object(dict_cfg)  # type: ignore
        if cfg.run_analysis:
            if analysis_cls is None:
                raise ValueError("Analysis class not provided!")
            a = analysis_cls(cfg)
            return a._analyze_wrapper()
        else:
            e = experiment_cls(cfg)
            return e.run()

    return main()



"""Hydra config resolvers."""

from omegaconf import OmegaConf
from inspect import isfunction

from experiment_lab.common.resolvers.list_resolvers import (
    dup,
    wrap_around,
    toggle,
    range,
    arange,
    linspace,
)
from experiment_lab.common.resolvers.object_resolvers import (
    get_class,
    get_object,
    get_method,
)


def register_resolvers() -> None:
    """Registers the resolvers imported to this file."""
    ignores = set(["register_resolvers", "isfunction"])

    resolvers = {
        k: v for k, v in globals().items() if k not in ignores and isfunction(v)
    }

    for k, v in resolvers.items():
        OmegaConf.register_new_resolver(k, v)



"""Python file for the post proessors"""

from typing import Any, Dict
import abc


class BasePostProcessor(abc.ABC):
    """The base post processor class."""

    def __init__(self) -> None:
        """The constructor for a generic post processor."""
        super().__init__()

    @abc.abstractmethod
    def process(self, result: Any) -> Dict[str, Any]:
        """Applies post processing on the aggregated result.

        Args:
            result (Any): The aggregated result.

        Returns:
            Any: The post processed result (publishable to wandb with names).
        """
        pass


class PassThroughPostProcessor(BasePostProcessor):
    """The pass through post processor."""

    def __init__(self) -> None:
        """The constructor for the pass through post processor."""
        super().__init__()

    def process(self, result: Any) -> Dict[str, Any]:
        """Returns the passed in result.

        Args:
            result (Any): The aggregated result.

        Returns:
            Any: The same result wrapped in a dict.
        """
        return {"result": result}



"""Python file containing the configs for monte carlo experiments."""

from dataclasses import dataclass, field
from typing import Any, Dict
from hydra.core.config_store import ConfigStore

from experiment_lab.core.base_config import BaseConfig


@dataclass
class MCConfig(BaseConfig):
    """The Monte Carlo dataclass."""

    n_samples: int = 10
    batch_size: int | None = None
    aggregate_every_batch: bool = False
    log_results: bool = False

    sampler: Dict[str, Any] = field(
        default_factory=lambda: {
            "_target_": "experiment_lab.experiments.monte_carlo.UniformSampler"
        }
    )

    aggregator: Dict[str, Any] = field(
        default_factory=lambda: {
            "_target_": "experiment_lab.experiments.monte_carlo.NpAggregator"
        }
    )

    sample_filter: Dict[str, Any] = field(
        default_factory=lambda: {
            "_target_": "experiment_lab.experiments.monte_carlo.PassThroughSampleFilter"
        }
    )

    post_processor: Dict[str, Any] = field(
        default_factory=lambda: {
            "_target_": "experiment_lab.experiments.monte_carlo.PassThroughPostProcessor"
        }
    )

    def __post_init__(self) -> None:
        """Does additional checks on loaded config."""
        super().__post_init__()
        assert self.n_samples >= 0, "Number of samples must be at least 0."
        assert (
            self.batch_size is None or self.batch_size > 0
        ), "Number of samples must be at least 1 or set to None"


def register_configs():
    """Registers the rl config."""
    cs = ConfigStore.instance()
    cs.store(name="mc_config", node=MCConfig)
